import bionty as bt
import pandas as pd
import pytest
from bionty._organism import OrganismNotSet
from bionty.models import DoesNotExist, InvalidArgument


def test_from_source():
    record = bt.Gene.from_source(symbol="BRCA2", organism="human")
    assert record.ensembl_gene_id == "ENSG00000139618"

    with pytest.raises(
        DoesNotExist, match=r"No record found in source for the given field value"
    ):
        bt.CellType.from_source(name="T-cellx")

    with pytest.raises(
        InvalidArgument,
        match=r"Only one field can be passed to generate records from source",
    ):
        bt.CellType.from_source(name="T cell", ontology_id="CL:0000084")

    with pytest.raises(
        InvalidArgument, match=r"No field passed to generate records from source"
    ):
        bt.CellType.from_source()


def test_get_source_record():
    from bionty._source import get_source_record

    source = get_source_record(bt.Phenotype, organism="human")
    assert source.organism == "human"

    source = get_source_record(bt.Phenotype)
    assert source.organism == "all"

    source = get_source_record(bt.Gene, organism="mouse")
    assert source.organism == "mouse"


def test_add_source():
    import wetlab as wl

    # pass a source record from another entity
    with pytest.raises(ValueError):
        wl.Compound.import_source()
    chebi_source = bt.Source.get(name="chebi", version="2024-07-27")
    new_source = wl.Compound.add_source(chebi_source)
    assert new_source.entity == "wetlab.Compound"
    assert new_source.name == "chebi"
    assert new_source.version == "2024-07-27"
    assert new_source.dataframe_artifact is not None
    public_ontology = wl.Compound.public()
    assert public_ontology.__class__.__name__ == "StaticReference"

    # pass a PublicOntology object
    public_ontology_ct_2024_05_15 = bt.base.CellType(source="cl", version="2024-05-15")
    new_source = bt.CellType.add_source(public_ontology_ct_2024_05_15)
    assert new_source.entity == "bionty.CellType"
    assert new_source.name == "cl"
    assert new_source.version == "2024-05-15"
    assert new_source.dataframe_artifact is not None

    # pass a source name and/ or version
    new_source = bt.Phenotype.add_source("oba")
    assert new_source.entity == "bionty.Phenotype"
    assert new_source.name == "oba"
    assert bt.Phenotype.add_source("oba", version=new_source.version) == new_source
    public_df = bt.Phenotype.public(source=new_source).to_dataframe()
    assert all(public_df.index.str.startswith("OBA:"))  # restrict_to_prefix


def test_base_gene_register_source_in_lamindb():
    bt.Organism.filter().delete(permanent=True)
    assert not bt.Source.filter(organism="rabbit").exists()
    gene = bt.base.Gene(source="ensembl", organism="rabbit", version="release-112")
    assert gene.to_dataframe().shape[0] > 10000
    gene.register_source_in_lamindb()
    assert bt.Organism.filter(name="rabbit").exists()
    rabbit_gene_source = bt.Source.get(organism="rabbit")
    assert rabbit_gene_source.entity == "bionty.Gene"
    assert rabbit_gene_source.name == "ensembl"
    assert rabbit_gene_source.version == "release-112"
    assert rabbit_gene_source.organism == "rabbit"


def test_import_source():
    # when adding a single record, it's parents are also added
    record = bt.Ethnicity.from_source(ontology_id="HANCESTRO:0005").save()
    parent = bt.Ethnicity.get(ontology_id="HANCESTRO:0004")
    assert parent in record.parents.filter().to_list()

    # bulk import should fill in gaps of missing parents
    parent.delete(permanent=True)
    bt.Ethnicity.import_source()
    parent = bt.Ethnicity.get(ontology_id="HANCESTRO:0004")
    assert parent in record.parents.filter().to_list()
    record = bt.Ethnicity.get("7RNCY3yC")
    assert record.parents.count() > 0
    # the source.in_db should be set to True since we imported all the records
    assert record.source.in_db is True

    # organism is required
    bt.settings._organism = None
    with pytest.raises(
        OrganismNotSet,
        match=r"CellMarker requires to specify a organism name via `organism=` or `bionty\.settings\.organism=`!",
    ):
        bt.CellMarker.import_source()
    bt.CellMarker.import_source(organism="mouse")
    assert bt.CellMarker.filter(organism__name="mouse").exists()

    # import source with stable_id
    bt.Gene.import_source(organism="saccharomyces cerevisiae")
    assert bt.Gene.filter(organism__name="saccharomyces cerevisiae").exists()

    # import organism data without 'parents' column (should not crash with KeyError)
    bt.Organism.import_source()
    assert bt.Organism.filter().count() > 0

    # strings are not supported (only Sources) and should raise a TypeError
    with pytest.raises(
        TypeError, match=r"import_source\(\) expects a `bt\.Source` object, not a str"
    ):
        bt.CellLine.import_source(source="depmap")

    # test that ICD Source does not crash
    source = bt.Disease.add_source(source="icd", version="icd-11-2023")
    bt.Disease.import_source(source=source)

    # import source with public df having extra columns than model fields
    bt.CellLine.import_source(source=bt.Source.get(name="depmap", version="2024-Q2"))
    assert bt.CellLine.filter(source__name="depmap").count() == 1959


def test_add_ontology_from_values():
    bt.Ethnicity.filter().delete(permanent=True)
    # .save() calls add_ontology() under the hood
    bt.Ethnicity.from_values(
        [
            "HANCESTRO:0597",
            "HANCESTRO:0006",
        ],
        field=bt.Ethnicity.ontology_id,
    ).save()
    record = bt.Ethnicity.get("7RNCY3yC")
    assert record.parents.count() > 0
    # the source.in_db should be set back to False since we deleted all records
    assert record.source.in_db is False


def test_add_custom_source():
    internal_source = bt.Source(
        entity="bionty.Gene",
        name="internal",
        version="0.0.1",
        organism="rabbit",
        description="internal gene reference",
    ).save()

    # without a dataframe artifact
    assert bt.Gene.public(source=bt.Source.get(name="internal")).to_dataframe().empty
    records = bt.Gene.from_values(
        ["ENSOCUG00000017195"],
        field=bt.Gene.ensembl_gene_id,
        source=internal_source,
    )
    assert len(records) == 0

    # with a dataframe artifact
    df = pd.DataFrame(
        {
            "ensembl_gene_id": ["ENSOCUG00000017195"],
            "symbol": ["SEL1L3"],
            "description": ["SEL1L family member 3"],
        }
    )
    internal_source = bt.Gene.add_source(internal_source, df=df)
    records = bt.Gene.from_values(
        ["ENSOCUG00000017195"],
        field=bt.Gene.ensembl_gene_id,
        source=internal_source,
    )
    assert len(records) == 1


def test_sync_public_sources():
    source_gene_latest = bt.Source.get(
        entity="bionty.Gene", name="ensembl", organism="mouse", currently_used=True
    )
    source_gene_release_111 = bt.Gene.add_source(
        source="ensembl", organism="mouse", version="release-111"
    )
    source_gene_release_111.currently_used = True
    # .save() updates currently_used of others to False
    source_gene_release_111.save()
    assert not bt.Source.get(source_gene_latest.uid).currently_used

    bt.Source.get(entity="bionty.CellType", name="cl", currently_used=True).delete(
        permanent=True
    )
    source_ct_2024_05_15 = bt.CellType.add_source(source="cl", version="2024-05-15")
    source_ct_2024_05_15.currently_used = True
    source_ct_2024_05_15.save()

    # update_currently_used=False
    source_gene_latest.delete(permanent=True)
    bt.core.sync_public_sources()
    source_gene_latest = bt.Source.get(
        entity="bionty.Gene", name="ensembl", organism="mouse", currently_used=True
    )
    assert source_gene_latest == source_gene_release_111
    source_cl_latest = bt.Source.get(
        entity="bionty.CellType", name="cl", currently_used=True
    )
    assert source_cl_latest == source_ct_2024_05_15

    # update_currently_used=True
    bt.core.sync_public_sources(update_currently_used=True)
    source_gene_latest = bt.Source.get(
        entity="bionty.Gene", name="ensembl", organism="mouse", currently_used=True
    )
    assert source_gene_latest != source_gene_release_111
    source_cl_latest = bt.Source.get(
        entity="bionty.CellType", name="cl", currently_used=True
    )
    assert source_cl_latest != source_ct_2024_05_15


def test_import_source_update_records():
    import lamindb as ln

    source1 = bt.CellType.add_source(source="cl", version="2022-08-16")
    source2 = bt.CellType.add_source(source="cl", version="2024-08-16")
    bt.CellType.import_source(source=source1)

    artifact = ln.Artifact(
        pd.DataFrame({"a": [1, 2, 3]}), key="test-upgrade-labels.parquet"
    ).save()
    record_wo_artifact_name = bt.CellType.get(ontology_id="CL:0000409").name
    record_w_artifact = bt.CellType.get(ontology_id="CL:0000003")
    artifact.cell_types.add(record_w_artifact)

    # records with artifacts should not be upgraded
    bt.CellType.import_source(source=source2, update_records=True)
    record_w_artifact = bt.CellType.get(ontology_id="CL:0000003")
    assert record_w_artifact.source == source1

    # records without artifacts should be upgraded
    record_wo_artifact = bt.CellType.get(ontology_id="CL:0000409")
    assert record_wo_artifact.source == source2
    assert record_wo_artifact.name != record_wo_artifact_name


def test_import_source_no_prefix_filter():
    efo_public = bt.base.ExperimentalFactor(version="3.78.0")
    efo_ontology_df = efo_public.to_pronto().to_df(include_id_prefixes=None)
    assert efo_ontology_df.shape[0] == 66971
    assert efo_ontology_df.index.str.startswith("EFO:").sum() == 18229
    source = bt.ExperimentalFactor.add_source(source="efo", df=efo_ontology_df)
    bt.ExperimentalFactor.import_source(source=source)
    assert bt.ExperimentalFactor.filter().count() == 66971



import bionty as bt


def test_public_synonym_mapping():
    bt_result = bt.Gene.public(organism="human").inspect(
        ["ABC1", "TNFRSF4"], field="symbol"
    )
    assert bt_result.synonyms_mapper == {"ABC1": "HEATR6"}

    bt_result = bt.Gene.public(organism="human").inspect(
        ["ABC1", "TNFRSF4"], field="symbol", inspect_synonyms=False
    )
    assert bt_result.synonyms_mapper == {}


def test_encode_uids():
    cell_type = bt.CellType(
        ontology_id="CL:0000084",
        _skip_validation=True,
    )
    assert cell_type.uid == "22LvKd01"

    organism = bt.Organism(
        ontology_id="NCBITaxon:9606",
        name="human",
        _skip_validation=True,
    )
    assert organism.uid == "1dpCL6Td"

    bt.settings.organism = "human"
    cell_marker = bt.CellMarker(
        name="test",
        organism=bt.settings.organism,
        _skip_validation=True,
    )
    assert cell_marker.uid == "2dZ52W9noUDK"

    gene = bt.Gene(
        ensembl_gene_id="ENSG00000081059",
        symbol="TCF7",
        organism=bt.settings.organism,  # required
        _skip_validation=True,
    )
    assert gene.uid == "7IkHKPl0ScQR"

    disease = bt.Source(
        entity="bionty.Disease",
        name="mondo",
        version="2023-04-04",
        organism="all",
        _skip_validation=True,
    )
    assert disease.uid == "Hgw08Vk3"

    phenotype = bt.Source(
        entity="bionty.Phenotype",
        name="hp",
        version="2023-06-17",
        organism="human",
        _skip_validation=True,
    )
    assert phenotype.uid == "451W7iJS"



# Generated by Django 5.1 on 2024-07-25 13:59

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0029_alter_cellline_previous_runs_and_more"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="PublicSource",
            new_name="Source",
        ),
        migrations.RenameField(
            model_name="cellline",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="cellmarker",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="celltype",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="developmentalstage",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="disease",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="ethnicity",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="experimentalfactor",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="gene",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="organism",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="pathway",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="phenotype",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="protein",
            old_name="public_source",
            new_name="source",
        ),
        migrations.RenameField(
            model_name="tissue",
            old_name="public_source",
            new_name="source",
        ),
        migrations.AddField(
            model_name="source",
            name="in_db",
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AddField(
            model_name="source",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="reference_of_sources", to="lamindb.artifact"
            ),
        ),
        migrations.AddField(
            model_name="source",
            name="df",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="reference_of_source",
                to="lamindb.artifact",
            ),
        ),
        migrations.AddField(
            model_name="cellmarker",
            name="description",
            field=models.TextField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="organism",
            name="description",
            field=models.TextField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="organism",
            name="synonyms",
            field=models.TextField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="protein",
            name="description",
            field=models.TextField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name="protein",
            name="name",
            field=models.CharField(
                db_index=True, default=None, max_length=256, null=True
            ),
        ),
        migrations.RenameField(
            model_name="artifactcellline",
            old_name="cell_line_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactcellmarker",
            old_name="cell_marker_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactcelltype",
            old_name="cell_type_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactdevelopmentalstage",
            old_name="developmental_stage_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactdisease",
            old_name="disease_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactethnicity",
            old_name="ethnicity_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactexperimentalfactor",
            old_name="experimental_factor",
            new_name="experimentalfactor",
        ),
        migrations.RenameField(
            model_name="artifactexperimentalfactor",
            old_name="experimental_factor_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactorganism",
            old_name="organism_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactpathway",
            old_name="pathway_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactphenotype",
            old_name="phenotype_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifactprotein",
            old_name="protein_ref_is_name",
            new_name="label_ref_is_name",
        ),
        migrations.RenameField(
            model_name="artifacttissue",
            old_name="tissue_ref_is_name",
            new_name="label_ref_is_name",
        ),
    ]



from __future__ import annotations

import importlib
import logging
from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from lamin_utils import logger
from lamin_utils._lookup import Lookup
from lamindb_setup.core import deprecated

from ._settings import check_datasetdir_exists, check_dynamicdir_exists, settings
from .dev._handle_sources import LAMINDB_INSTANCE_LOADED
from .dev._io import s3_bionty_assets, url_download

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from bionty.base._ontology import Ontology

    from .dev import InspectResult


def encode_filenames(
    organism: str, name: str, version: str, entity: str
) -> tuple[str, str]:
    """Encode names of the cached files."""
    # Paths are often passed as `bionty.Entity` but we only need the entity here
    entity_name = entity.split(".")[-1] if "." in entity else entity
    parquet_filename = f"df_{organism}__{name}__{version}__{entity_name}.parquet"
    ontology_filename = (
        f"ontology_{organism}__{name}__{version}__{entity_name}".replace(" ", "_")
    )

    return parquet_filename, ontology_filename


class PublicOntology:
    """PublicOntology object."""

    def __init__(
        self,
        source: str | None = None,
        version: str | None = None,
        organism: str | None = None,
        *,
        include_id_prefixes: dict[str, list[str]] | None = None,
        include_rel: str | None = None,
        entity: str | None = None,
        ols_supported: bool = True,
        filter_prefix: bool = True,
    ):
        self._entity = entity or self.__class__.__name__
        self._ols_supported = ols_supported
        self._filter_prefix = filter_prefix

        # search in all available sources to get url
        try:
            self._fetch_sources()
            self._source_dict = self._match_sources(
                self._all_sources,
                name=source,
                version=version,
                organism=organism,
            )

        except (ValueError, KeyError) as e:
            if LAMINDB_INSTANCE_LOADED:
                # to support StaticSource in lamindb
                self._source_dict = {}
                pass
            else:
                raise e

        self._organism = self._source_dict.get("organism") or organism
        self._source = self._source_dict.get("name") or source
        self._version = self._source_dict.get("version") or version

        self._get_url()
        self._set_file_paths()
        self.include_id_prefixes = include_id_prefixes
        self.include_rel = include_rel

        # df is only read into memory at the init to improve performance
        df = self._load_df()
        # self._df has no index
        if not df.empty and df.index.name is not None:
            df = df.reset_index()
        self._df: pd.DataFrame = df

        # set column names/fields as attributes
        for col_name in self._df.columns:
            try:
                setattr(self, col_name, PublicOntologyField(self, col_name))
            # Some fields of an ontology (e.g. Gene) are not PublicOntology class attributes and must be skipped.
            except AttributeError:
                pass

    def __repr__(self) -> str:
        # fmt: off
        representation = (
            f"PublicOntology\n"
            f"Entity: {self._entity}\n"
            f"Organism: {self.organism}\n"
            f"Source: {self.source}, {self.version}\n"
            f"#terms: {self._df.shape[0] if hasattr(self, '_df') else ''}\n\n"
        )
        # fmt: on
        return representation

    @property
    def organism(self):
        """The `name` of `Organism`."""
        return self._organism

    @property
    def source(self):
        """Name of the source."""
        return self._source

    @property
    def version(self):
        """Version of the source."""
        return self._version

    @property
    def fields(self) -> set:
        """All PublicOntology entity fields."""
        blacklist = {"include_id_prefixes", "include_rel"}
        fields = {
            field
            for field in vars(self)
            if not callable(getattr(self, field)) and not field.startswith("_")
        }
        return fields - blacklist

    def _download_ontology_file(self, localpath: Path, url: str) -> None:
        """Download ontology source file to _local_ontology_path."""
        if not localpath.exists():
            logger.info(f"downloading {self._entity} ontology source file...")
            if not url:
                raise ValueError(
                    f"No source url is available for {self._entity} ontology."
                )
            self._url_download(url, localpath)

    def _fetch_sources(self) -> None:
        from .dev._handle_sources import parse_sources_yaml

        self._all_sources = parse_sources_yaml(url_pattern=True)

    def _match_sources(
        self,
        ref_sources: pd.DataFrame,
        name: str | None = None,
        version: str | None = None,
        organism: str | None = None,
    ) -> dict[str, str]:
        """Match a source record based on passed organism, name and version."""
        # Build filter conditions from non-None parameters
        conditions = {}
        if organism is not None:
            conditions["organism"] = organism
        if name is not None:
            conditions["name"] = name

        # If no parameters provided, use the first source
        if not conditions:
            rows = ref_sources
        else:
            # Filter dataframe using all conditions
            query = True
            for col, value in conditions.items():
                query = query & (ref_sources[col] == value)
            rows = ref_sources[query]

        # For ensembl, need to filter by entity
        if not rows.empty and rows.head(1)["name"].values[0] == "ensembl":
            rows = rows[rows["entity"] == self._entity.split(".")[-1]]

        row = rows.head(1)

        # If no records matched
        if row.empty:
            url = None
            ontology_version = None
            if self._ols_supported:
                # try to get the ontology url
                if name and name not in ref_sources["name"].values:
                    from ._ontology_url import get_ontology_url

                    try:
                        url, ontology_version = get_ontology_url(
                            prefix=name, version=version
                        )
                        meta_dict = {
                            "name": name,
                            "version": ontology_version,
                            "organism": organism or "all",
                            "url": url,
                            "description": f"{name} Ontology",
                            "source_website": f"https://www.ebi.ac.uk/ols/ontologies/{name}",
                        }

                        return meta_dict
                    except Exception:
                        pass
            if url is None:
                raise ValueError(
                    f"No source is available with {conditions}\n"
                    "Check `.display_sources()`"
                )

        # Build result and replace version placeholder in URL
        meta_dict = row.iloc[0].to_dict()
        ver = version or meta_dict.get("version")
        meta_dict["version"] = ver
        meta_dict["url"] = meta_dict["url"].replace("{version}", ver)
        meta_dict.pop("entity", None)  # remove entity from the dict
        return meta_dict

    @check_dynamicdir_exists
    def _url_download(self, url: str, localpath: Path) -> None:
        """Download file from url to dynamicdir _local_ontology_path."""
        # Try to download from s3://bionty-assets
        s3_bionty_assets(
            filename=self._ontology_filename,  # type: ignore
            assets_base_url="s3://bionty-assets",
            localpath=localpath,
        )

        # If the file is not available, download from the url
        if not localpath.exists():
            logger.info(f"downloading {self._entity} source file from: {url}")
            _ = url_download(url, localpath)

    def _get_url(self):
        """Get the url of the source."""
        self._url: str = self._source_dict.get("url", "")

    @check_datasetdir_exists
    def _set_file_paths(self) -> None:
        """Sets local file paths."""
        # parquet file name, ontology source file name
        self._parquet_filename, self._ontology_filename = encode_filenames(
            organism=self.organism,
            name=self.source,
            version=self.version,
            entity=self._entity,
        )
        self._local_parquet_path: Path = settings.dynamicdir / self._parquet_filename

        # user provide reference table as the url in parquet format
        if self._url.endswith(".parquet"):
            # no local ontology source file
            self._local_ontology_path = None  # type:ignore
            if not self._url.startswith("s3://bionty-assets/"):
                # will not try to download the parquet file from s3://bionty-assets
                self._parquet_filename = None  # type:ignore
        else:
            self._local_ontology_path = settings.dynamicdir / self._ontology_filename

        # ontology source not present in the sources.yaml file
        # these entities don't have ontology files
        if not self._url and not self._ols_supported:
            self._local_ontology_path = None

    def _get_default_field(self, field: PublicOntologyField | str | None = None) -> str:
        """Default to name field."""
        if field is None:
            if "name" in self._df.columns:
                field = "name"
            elif "symbol" in self._df.columns:
                field = "symbol"
            else:
                raise ValueError("Please specify a field!")
        field = str(field)
        if field not in self._df.columns:
            raise AssertionError(f"No {field} column exists!")
        return field

    def _load_df(self) -> pd.DataFrame:
        if self._parquet_filename is None:
            self._url_download(self._url, self._local_parquet_path)
        else:
            s3_bionty_assets(
                filename=self._parquet_filename,
                assets_base_url="s3://bionty-assets",
                localpath=self._local_parquet_path,
            )
        # If download is not possible, write a parquet file of the ontology df
        if not self._url.endswith(".parquet"):
            if not self._local_parquet_path.exists():
                pronto = self.to_pronto(mute=True)
                if pronto is not None:
                    df = pronto.to_df(
                        source=self.source,
                        include_id_prefixes=self.include_id_prefixes,
                        include_rel=self.include_rel,
                    )
                    df.to_parquet(self._local_parquet_path)

        if self._local_parquet_path.exists():
            # Loading the parquet file resets the index
            return pd.read_parquet(self._local_parquet_path)
        return pd.DataFrame()

    def to_pronto(self, mute: bool = False) -> Ontology:  # type:ignore
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        if importlib.util.find_spec("pronto") is None:  # type:ignore
            raise ImportError(
                "pronto package is not installed. Please install it using: pip install pronto."
            )

        from ._ontology import Ontology

        if self._local_ontology_path is None:
            if not mute:
                logger.error(f"{self._entity} has no Pronto Ontology object!")
        else:
            self._download_ontology_file(
                localpath=self._local_ontology_path,
                url=self._url,
            )
            return Ontology(handle=self._local_ontology_path)

    def to_dataframe(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Example::

            import bionty.base as bt_base

            bt_base.Gene().to_dataframe()
        """
        if "ontology_id" in self._df.columns:
            if self._filter_prefix:
                # Filter ontology_id by source prefix
                filtered_df = self._df[
                    self._df["ontology_id"]
                    .str.upper()
                    .str.startswith(f"{self._source.upper()}:")
                ].set_index("ontology_id")
                if not filtered_df.empty:
                    return filtered_df
            return self._df.set_index("ontology_id")
        else:
            return self._df

    @deprecated("to_dataframe")
    def df(self) -> pd.DataFrame:
        return self.to_dataframe()

    def validate(
        self,
        values: Iterable,
        field: PublicOntologyField,
        *,
        mute: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Validate a list of values against a field of entity reference.

        Args:
            values: Identifiers that will be checked against the field.
            field: The PublicOntologyField of the ontology to compare against.
                Examples are 'ontology_id' to map against the source ID
                or 'name' to map against the ontologies field names.
            mute: Whether to suppress logging. Defaults to False.
            kwargs: Used for backwards compatibility and return types.

        Returns:
            A boolean array indicating compliance validation.

        Example::

            import bionty.base as bt_base

            public = bt_base.Gene()
            gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            public.validate(gene_symbols, field=public.symbol)
        """
        from lamin_utils._inspect import validate

        if isinstance(values, str):
            values = [values]

        # in bionty-base passed django fields do not resolve properly to a string
        if str(field).startswith("FieldAttr"):
            field_str = str(field).split(".")[-1][:-1]
        else:
            field_str = str(field)
        field_values = self._df[str(field_str)]

        return validate(
            identifiers=values,
            field_values=field_values,
            case_sensitive=True,
            mute=mute,
            **kwargs,
        )

    def inspect(
        self,
        values: Iterable,
        field: PublicOntologyField,
        *,
        mute: bool = False,
        **kwargs,
    ) -> InspectResult:
        """Inspect a list of values against a field of entity reference.

        Args:
            values: Identifiers that will be checked against the field.
            field: The PublicOntologyField of the ontology to compare against.
                Examples are 'ontology_id' to map against the source ID
                or 'name' to map against the ontologies field names.
            return_df: Whether to return a Pandas DataFrame.
            mute: Whether to suppress logging. Defaults to False.
            kwargs: Used for backwards compatibility and return types.

        Returns:
            - A Dictionary of "validated" and "not_validated" identifiers
            - If `return_df`: A DataFrame indexed by identifiers with a boolean
                `__validated__` column indicating compliance validation.

        Example::

            import bionty.base as bt_base

            public = bt_base.Gene()
            gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            public.inspect(gene_symbols, field=public.symbol)
        """
        from lamin_utils._inspect import inspect

        if isinstance(values, str):
            values = [values]

        return inspect(
            df=self._df,
            identifiers=values,
            field=str(field),
            mute=mute,
            **kwargs,
        )

    # unfortunately, the doc string here is duplicated with ORM.standardize
    def standardize(
        self,
        values: Iterable,
        field: PublicOntologyField | str | None = None,
        *,
        return_field: str = None,
        return_mapper: bool = False,
        case_sensitive: bool = False,
        mute: bool = False,
        keep: Literal["first", "last", False] = "first",
        synonyms_field: PublicOntologyField | str = "synonyms",
    ) -> dict[str, str] | list[str]:
        """Convert into standardized names.

        Args:
            values: `Iterable` Synonyms that will be standardized.
            field: `Optional[str]` The field representing the standardized names.
            return_field: `Optional[str]` The field to return. Defaults to field.
            return_mapper: `bool = False` If `True`, returns `{input_synonym1:
                standardized_name1}`.
            case_sensitive: `bool = False` Whether the mapping is case sensitive.
            keep: {'first', 'last', False}, default 'first'.
                When a synonym maps to multiple standardized values, determines
                which duplicates to mark as `pandas.DataFrame.duplicated`.

                - "first": returns the first mapped standardized value
                - "last": returns the last mapped standardized value
                - False: returns all mapped standardized value
            mute: Whether to mute logging. Defaults to False.
            synonyms_field: `str = "synonyms"` A field containing the concatenated synonyms.

        Returns:
            If `return_mapper` is `False`: a list of standardized names. Otherwise,
            a dictionary of mapped values with mappable synonyms as keys and
            standardized names as values.

        Example::

            import bionty.base as bt_base

            public = bt_base.Gene()
            gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            standardized_symbols = public.standardize(gene_symbols, public.symbol)
        """
        from lamin_utils._standardize import standardize as map_synonyms

        if isinstance(values, str):
            values = [values]

        return map_synonyms(
            df=self._df,
            identifiers=values,
            field=self._get_default_field(field),
            return_field=self._get_default_field(return_field),
            return_mapper=return_mapper,
            case_sensitive=case_sensitive,
            mute=mute,
            keep=keep,
            synonyms_field=str(synonyms_field),
        )

    def lookup(self, field: PublicOntologyField | str | None = None) -> tuple:
        """An auto-complete object for a PublicOntology field.

        Args:
            field: The field to lookup the values for.
                Defaults to 'name'.

        Returns:
            A NamedTuple of lookup information of the field values.

        Example::

            import bionty.base as bt_base

            lookup = bt_base.CellType().lookup()
            lookup.cd103_positive_dendritic_cell
            lookup_dict = lookup.dict()
            lookup['CD103-positive dendritic cell']
        """
        return Lookup(
            df=self._df,
            field=self._get_default_field(field),
            tuple_name=self._entity,
            prefix="bt",
        ).lookup()

    def search(
        self,
        string: str,
        *,
        field: PublicOntologyField | str | list[PublicOntologyField | str] = None,
        limit: int | None = None,
        case_sensitive: bool = False,
    ):
        """Search a given string against a PublicOntology field or fields.

        Args:
            string: The input string to match against the field values.
            field: The PublicOntologyField or several fileds of the ontology
                the input string is matching against. Search all fields containing strings by default.
            limit: Maximum amount of top results to return. If None, return all results.
            case_sensitive: Whether the match is case sensitive.

        Returns:
            Ranked search results.

        Example::

            import bionty.base as bt_base

            public = bt_base.CellType()
            public.search("gamma delta T cell")
        """
        from lamin_utils._search import search

        if isinstance(field, PublicOntologyField):
            field = field.name
        elif field is not None and not isinstance(field, str):
            field = [f.name if isinstance(f, PublicOntologyField) else f for f in field]

        result = search(
            df=self._df,
            string=string,
            field=field,
            limit=limit,
            case_sensitive=case_sensitive,
        )
        if "ontology_id" in result.columns:
            result = result.set_index("ontology_id")
        return result

    def diff(
        self, compare_to: PublicOntology, **kwargs
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Determines a diff between two PublicOntology objects' ontologies.

        Args:
            compare_to: PublicOntology object that must be of the same class as the calling object.
            kwargs: Are passed to pd.DataFrame.compare()

        Returns:
            A tuple of two DataFrames:
            1. New entries.
            2. A pd.DataFrame.compare result which denotes all changes in `self` and `other`.

        Example::

            import bionty.base as bt_base

            public_1 = bt_base.Disease(source="mondo", version="2023-04-04")
            public_2 = bt_base.Disease(source="mondo", version="2023-04-04")
            new_entries, modified_entries = public_1.diff(public_2)
            print(new_entries.head())
            print(modified_entries.head())
        """
        if type(self) is not type(compare_to):
            raise ValueError("Both PublicOntology objects must be of the same class.")

        if not self.source == compare_to.source:
            raise ValueError("Both PublicOntology objects must use the same source.")

        if self.version == compare_to.version:
            raise ValueError("The versions of the PublicOntology objects must differ.")

        # The 'parents' column (among potentially others) contain Numpy array values.
        # We transform them to tuples to determine the diff.
        def _convert_arrays_to_tuples(arr):  # pragma: no cover
            if isinstance(arr, np.ndarray):
                return tuple(arr)
            else:
                return arr

        for bt_obj in [self, compare_to]:
            for column in bt_obj.to_dataframe().columns:
                if any(
                    isinstance(val, np.ndarray) for val in bt_obj.to_dataframe()[column]
                ):
                    bt_obj._df[column] = bt_obj.to_dataframe()[column].apply(
                        _convert_arrays_to_tuples
                    )

        # New entries
        new_entries = pd.concat(
            [self.to_dataframe(), compare_to.to_dataframe()]
        ).drop_duplicates(keep=False)

        # Changes in existing entries
        common_index = self.to_dataframe().index.intersection(
            compare_to.to_dataframe().index
        )
        self_df_common = self.to_dataframe().loc[common_index]
        compare_to_df_common = compare_to.to_dataframe().loc[common_index]
        modified_entries = self_df_common.compare(compare_to_df_common, **kwargs)

        logging.info(f"{len(new_entries)} new entries were added.")
        logging.info(f"{len(modified_entries)} entries were modified.")

        return new_entries, modified_entries


class InvalidParamError(Exception):
    """Custom exception for PublicOntology parameter validation errors."""

    pass


class PublicOntologyField:
    """Field of a PublicOntology model."""

    def __init__(self, parent: PublicOntology, name: str):
        self.parent = parent
        self.name = name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name



# Generated by Django 5.1.12 on 2025-10-01 14:39

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0060_alter_cellline_is_locked_alter_cellmarker_is_locked_and_more"),
    ]

    operations = []  # type: ignore



# Generated by Django 5.2 on 2024-09-09 11:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0038_alter_artifactcellline_created_by_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cellline",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="bionty.source",
            ),
        ),
    ]



import bionty as bt
import pytest
from bionty._organism import OrganismNotSet


def test_from_values_organism():
    bt.settings._organism = None
    with pytest.raises(OrganismNotSet):
        bt.Gene.from_values(["ABC1"], bt.Gene.symbol)
    # no organism is needed if the values are ensembl gene ids
    result = bt.Gene.from_values(["ENSG00000068097"], bt.Gene.ensembl_gene_id)
    assert len(result) == 1
    result = bt.Gene.from_values(
        ["ENSMUSG00000102862", "ENSMUSG00000084826"], field=bt.Gene.ensembl_gene_id
    )
    assert len(result) == 2
    result = bt.Gene.from_values(
        ["HRA1", "ETS1-1"], field=bt.Gene.stable_id, organism="saccharomyces cerevisiae"
    )
    assert len(result) == 2

    bt.settings.organism = "human"
    values = ["ABC1"]
    standardized_values = bt.Gene.public().standardize(values)
    records = bt.Gene.from_values(standardized_values, bt.Gene.symbol)
    assert records[0].ensembl_gene_id == "ENSG00000068097"

    # TODO: Gene.public() should raise error if organism is not provided
    standardized_values = bt.Gene.public(organism="mouse").standardize(values)
    records = bt.Gene.from_values(standardized_values, bt.Gene.symbol, organism="mouse")
    assert records[0].ensembl_gene_id == "ENSMUSG00000015243"

    # clean up
    bt.settings._organism = None
    bt.Organism.filter().delete(permanent=True)
    bt.Gene.filter().delete(permanent=True)



"""Developer API.

.. autosummary::
   :toctree: .

   BioRecord
   StaticReference
   Settings
   sync_public_sources
"""

from lamindb_setup._check_setup import _check_instance_setup

_check_instance_setup(from_module="bionty")

from bionty.models import BioRecord, StaticReference

from ._add_ontology import add_ontology
from ._settings import Settings
from ._source import sync_public_sources

# backward-compat
sync_all_sources_to_latest = sync_public_sources



# Generated by Django 5.0.7 on 2024-07-30 13:23

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0032_rename_source_name_source_description_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifactcellline",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_cell_line",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="cellline",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.cellline",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactcellline",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_cell_marker",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="cellmarker",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.cellmarker",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactcellmarker",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_cell_type",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="celltype",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.celltype",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactcelltype",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_developmental_stage",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="developmentalstage",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.developmentalstage",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactdevelopmentalstage",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_disease",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="disease",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.disease",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactdisease",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_ethnicity",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="ethnicity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.ethnicity",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactethnicity",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_experimental_factor",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="experimentalfactor",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.experimentalfactor",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactexperimentalfactor",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_gene",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactgene",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="gene",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.gene",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_organism",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactorganism",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="organism",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.organism",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_pathway",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactpathway",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="pathway",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.pathway",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_phenotype",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactphenotype",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="phenotype",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.phenotype",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_protein",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifactprotein",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="protein",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.protein",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_tissue",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifacttissue",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="tissue",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="bionty.tissue",
            ),
        ),
        migrations.RenameField(
            model_name="cellline",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="cellmarker",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="celltype",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="developmentalstage",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="disease",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="ethnicity",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="experimentalfactor",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="gene",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="organism",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="pathway",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="phenotype",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="protein",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="source",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
        migrations.RenameField(
            model_name="tissue",
            old_name="previous_runs",
            new_name="_previous_runs",
        ),
    ]



import tempfile
from pathlib import Path

import pytest
from bionty.base.dev._handle_sources import (
    parse_currently_used_sources,
    parse_sources_yaml,
)


@pytest.fixture(scope="function")
def versions_yaml_replica():
    input_file_content = """
    version: "0.3.0"
    Organism:
      ensembl:
        vertebrates:
          latest-version: release-112
          url: https://ftp.ensembl.org/pub/{version}/species_EnsemblVertebrates.txt
        name: Ensembl
        website: https://www.ensembl.org/index.html
    Gene:
      ensembl:
        human:
          latest-version: release-112
          url: s3://bionty-assets/df_human__ensembl__{version}__Gene.parquet
        mouse:
          latest-version: release-112
          url: s3://bionty-assets/df_mouse__ensembl__{version}__Gene.parquet
        name: Ensembl
        website: https://www.ensembl.org/index.html
    CellType:
      cl:
        all:
          latest-version: 2024-08-16
          url: http://purl.obolibrary.org/obo/cl/releases/{version}/cl.owl
        name: Cell Ontology
        website: https://obophenotype.github.io/cell-ontology
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        f.write(input_file_content)
        f.flush()
        yield f.name

    Path(f.name).unlink()


def test_parse_versions_yaml(versions_yaml_replica):
    parsed_df = parse_sources_yaml(versions_yaml_replica)
    assert parsed_df.shape == (4, 7)
    assert all(parsed_df["entity"].values == ["Organism", "Gene", "Gene", "CellType"])
    assert all(parsed_df["organism"].values == ["vertebrates", "human", "mouse", "all"])
    assert all(parsed_df["name"].values == ["ensembl", "ensembl", "ensembl", "cl"])


def test_parse_current_versions(versions_yaml_replica):
    expected = {
        "Organism": {"vertebrates": {"ensembl": "release-112"}},
        "Gene": {
            "human": {"ensembl": "release-112"},
            "mouse": {"ensembl": "release-112"},
        },
        "CellType": {"all": {"cl": "2024-08-16"}},
    }

    assert parse_currently_used_sources(versions_yaml_replica) == expected



"""UIDs.

Entity-related generators:

.. autosummary::
   :toctree: .

   gene
   protein
   cellmarker
   ontology
   source

"""

import hashlib
import secrets
import string


def base62(n_char: int) -> str:
    """Random Base62 string."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def encode_base62(s: str) -> str:
    from lamin_utils._base62 import encodebytes

    return encodebytes(hashlib.md5(s.encode()).digest())


def hash_id(input_id: str | None = None, *, n_char: int) -> str:
    if input_id is None:
        return base62(n_char=n_char)
    else:
        return encode_base62(input_id)[:n_char]


def gene(input_id: str | None = None) -> str:
    """12 base62."""
    return hash_id(input_id, n_char=12)


def protein(input_id: str | None = None) -> str:
    """12 base62."""
    return hash_id(input_id, n_char=12)


def cellmarker(input_id: str | None = None) -> str:
    """12 base62."""
    return hash_id(input_id, n_char=12)


def ontology(input_id: str | None = None):
    """8 base62."""
    return hash_id(input_id, n_char=8)


def source(input_id: str | None = None):
    """8 base62."""
    return hash_id(input_id, n_char=8)


def encode_uid(registry: type, kwargs: dict):
    """The type passed needs to be a subclass of BioRecord."""
    from lamindb.models import SQLRecord

    from . import ids

    if kwargs.get("uid") is not None:
        # if uid is passed, no encoding is needed
        return kwargs
    name = registry.__name__.lower()
    if hasattr(registry, "organism_id"):
        organism = kwargs.get("organism")
        if organism is None:
            if kwargs.get("organism_id") is not None:
                from .models import Organism

                organism = Organism.get(kwargs.get("organism_id")).name
        elif isinstance(organism, SQLRecord):
            organism = organism.name
    else:
        organism = ""

    if hasattr(registry, "_ontology_id_field"):
        ontology_id_field = registry._ontology_id_field
    else:
        ontology_id_field = "ontology_id"
    if hasattr(registry, "_name_field"):
        name_field = registry._name_field
    else:
        name_field = "name"

    str_to_encode = None
    if name == "source":
        str_to_encode = f"{kwargs.get('entity', '')}{kwargs.get('name', '')}{kwargs.get('organism', '')}{kwargs.get('version', '')}"
    elif name == "gene":  # gene has multiple id fields
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("stable_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(
                f"must provide {ontology_id_field}, stable_id or {name_field}"
            )
    else:
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(f"must provide {ontology_id_field} or {name_field}")

    if str_to_encode is not None and len(str_to_encode) > 0:
        try:
            id_encoder = getattr(ids, name)
        except Exception:
            if ontology_id_field == "ontology_id":
                id_encoder = ids.ontology
            else:
                return kwargs
        kwargs["uid"] = id_encoder(str_to_encode)
    return kwargs


def encode_uid_for_hub(registry_name: str, registry_schema_json: dict, kwargs: dict):
    """Encode the uid for the hub.

    Note that `organism` record must be passed in kwargs instead of `organism_id`.
    """
    from . import ids

    if kwargs.get("uid") is not None:
        # if uid is passed, no encoding is needed
        return kwargs
    name = registry_name.lower()
    # here we need to pass the organism record, not organism_id
    organism = kwargs.get("organism", "")
    if organism:
        organism = organism.get("name", "")

    # default to ontology_id
    ontology_id_field = registry_schema_json.get("_ontology_id_field", "ontology_id")
    name_field = registry_schema_json.get("_name_field", "name")

    str_to_encode = None
    if name == "source":
        str_to_encode = f"{kwargs.get('entity', '')}{kwargs.get('name', '')}{kwargs.get('organism', '')}{kwargs.get('version', '')}"
    elif name == "gene":  # gene has multiple id fields
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("stable_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(
                f"must provide {ontology_id_field}, stable_id or {name_field}"
            )
    else:
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(f"must provide {ontology_id_field} or {name_field}")

    if str_to_encode is not None and len(str_to_encode) > 0:
        try:
            id_encoder = getattr(ids, name)
        except Exception:
            if ontology_id_field == "ontology_id":
                id_encoder = ids.ontology
            else:
                return kwargs
        kwargs["uid"] = id_encoder(str_to_encode)
    return kwargs



# Generated by Django 5.2 on 2025-06-01 05:27

import django.db.models.deletion
import django.db.models.functions.datetime
import lamindb.base.fields
import lamindb.base.users
import lamindb.models.run
import lamindb.models.sqlrecord
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0052_rename__branch_code_cellline_branch_and_more"),
        ("lamindb", "0103_remove_writelog_migration_state_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="RecordCellLine",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "cellline",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.cellline",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordcellline",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_cell_line",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "cellline", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="cellline",
            name="records",
            field=models.ManyToManyField(
                related_name="cell_lines",
                through="bionty.RecordCellLine",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordCellMarker",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "cellmarker",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.cellmarker",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordcellmarker",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_cell_marker",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "cellmarker", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="cellmarker",
            name="records",
            field=models.ManyToManyField(
                related_name="cell_markers",
                through="bionty.RecordCellMarker",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordCellType",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "celltype",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.celltype",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordcelltype",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_cell_type",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "celltype", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="celltype",
            name="records",
            field=models.ManyToManyField(
                related_name="cell_types",
                through="bionty.RecordCellType",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordDevelopmentalStage",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "developmentalstage",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.developmentalstage",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recorddevelopmentalstage",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_developmental_stage",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "developmentalstage", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="developmentalstage",
            name="records",
            field=models.ManyToManyField(
                related_name="developmental_stages",
                through="bionty.RecordDevelopmentalStage",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordDisease",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "disease",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.disease",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recorddisease",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_disease",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "disease", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="disease",
            name="records",
            field=models.ManyToManyField(
                related_name="diseases",
                through="bionty.RecordDisease",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordEthnicity",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "ethnicity",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.ethnicity",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordethnicity",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_ethnicity",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "ethnicity", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="ethnicity",
            name="records",
            field=models.ManyToManyField(
                related_name="ethnicities",
                through="bionty.RecordEthnicity",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordExperimentalFactor",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "experimentalfactor",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.experimentalfactor",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordexperimentalfactor",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_experimental_factor",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "experimentalfactor", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="experimentalfactor",
            name="records",
            field=models.ManyToManyField(
                related_name="experimental_factors",
                through="bionty.RecordExperimentalFactor",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordGene",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordgene",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "gene",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.gene",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_gene",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "gene", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="gene",
            name="records",
            field=models.ManyToManyField(
                related_name="genes", through="bionty.RecordGene", to="lamindb.record"
            ),
        ),
        migrations.CreateModel(
            name="RecordOrganism",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordorganism",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "organism",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.organism",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_organism",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "organism", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="organism",
            name="records",
            field=models.ManyToManyField(
                related_name="organisms",
                through="bionty.RecordOrganism",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordPathway",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordpathway",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "pathway",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.pathway",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_pathway",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "pathway", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="pathway",
            name="records",
            field=models.ManyToManyField(
                related_name="pathways",
                through="bionty.RecordPathway",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordPhenotype",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordphenotype",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "phenotype",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.phenotype",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_phenotype",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "phenotype", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="phenotype",
            name="records",
            field=models.ManyToManyField(
                related_name="phenotypes",
                through="bionty.RecordPhenotype",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordProtein",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordprotein",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "protein",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.protein",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_protein",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "protein", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="protein",
            name="records",
            field=models.ManyToManyField(
                related_name="proteins",
                through="bionty.RecordProtein",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordTissue",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_recordtissue",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_tissue",
                        to="lamindb.record",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "tissue",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="bionty.tissue",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "tissue", "feature")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="tissue",
            name="records",
            field=models.ManyToManyField(
                related_name="tissues",
                through="bionty.RecordTissue",
                to="lamindb.record",
            ),
        ),
    ]



import bionty.base as bt_base
import pandas as pd


def test_efo_experimental_factor_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "EFO:1002048",
            "EFO:1002050",
            "EFO:1002047",
            "EFO:1002049",
            "This readout does not exist",
        ]
    )

    ro = bt_base.ExperimentalFactor(source="efo")
    inspected_df = ro.inspect(df.index, ro.ontology_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_efo_shape():
    """We observed issues with new EFO versions not including all records."""
    # 3.78.0 is the latest version where had initially observed this issue
    # If this works well, we may unpin the fixed version
    assert bt_base.ExperimentalFactor(version="3.78.0").to_dataframe().shape[0] > 18000



from __future__ import annotations

from typing import TYPE_CHECKING

from .uids import encode_uid

if TYPE_CHECKING:
    from types import ModuleType

    from .models import BioRecord


def lookup2kwargs(record: BioRecord, *args, **kwargs) -> dict:
    """Pass bionty search/lookup results."""
    from ._organism import create_or_get_organism_record
    from ._source import get_source_record

    arg = args[0]
    if isinstance(arg, tuple):
        bionty_kwargs = arg._asdict()  # type:ignore
    else:
        bionty_kwargs = arg[0]._asdict()

    if len(bionty_kwargs) > 0:
        # add organism and source
        organism_record = create_or_get_organism_record(
            registry=record.__class__, organism=kwargs.get("organism")
        )
        if organism_record is not None:
            bionty_kwargs["organism"] = organism_record
        bionty_kwargs["source"] = get_source_record(
            registry=record.__class__,
            organism=organism_record,
            source=kwargs.get("source"),
        )

        model_field_names = {i.name for i in record._meta.fields}
        model_field_names.add("parents")
        bionty_kwargs = {
            k: v for k, v in bionty_kwargs.items() if k in model_field_names
        }
    return encode_uid(registry=record.__class__, kwargs=bionty_kwargs)


def list_biorecord_models(schema_module: ModuleType):
    """List all BioRecord models in a given schema module."""
    import inspect

    import lamindb as ln  # needed here

    from .models import BioRecord

    return [
        attr
        for attr in dir(schema_module.models)
        if inspect.isclass(getattr(schema_module.models, attr))
        and issubclass(getattr(schema_module.models, attr), BioRecord)
    ]
